from .base_data_model import BaseDataModel
from .data_filter import DataFilter
from .data_model import DataModel
from .base_internal import BaseInternal
from .base_redis_internal import BaseRedisInternal
from .redis_data_model import RedisDataModel


__all__ = [
    'BaseDataModel',
    'DataFilter',
    'DataModel',
    'BaseInternal',
    'BaseRedisInternal',
    'RedisDataModel',
]



import yaml
import os
import aiofiles

from typing import List

from eyja.db.rethinkdb import RethinkDBConfig
from eyja.db.redis import RedisConfig

from .config import Config


class ConfigHub:
    _config: Config

    @classmethod
    async def load_from_file(cls, config_file = os.environ.get('CONFIG', './config/local.yml')):
        async with aiofiles.open(config_file, 'r') as fp:
            config_data = await fp.read()

        cls.load(config_data)

    @classmethod
    def load(cls, data: str):
        config_data = yaml.load(data, Loader=yaml.FullLoader)
        cls._config = Config(**config_data)

    @classmethod
    def rethinkdb(cls) -> List[RethinkDBConfig]:
        return cls._config.rethinkdb

    @classmethod
    def redis(cls) -> List[RedisConfig]:
        return cls._config.redis



from typing import List, Union

from eyja.db.hub import DataHub
from eyja.utils import now

from .base_data_model import BaseDataModel
from .data_filter import DataFilter


class DataModel(BaseDataModel):
    def __init__(__pydantic_self__, **data) -> None:
        if 'object_id' not in data:
            data.setdefault('object_id', None)
            data.setdefault('created_at', now())
        data.setdefault('updated_at', now())
        super().__init__(**data)

    async def save(__pydantic_self__) -> None:
        __pydantic_self__.updated_at = now()
        await DataHub.save(__pydantic_self__)

    async def delete(__pydantic_self__) -> None:
        await DataHub.delete(__pydantic_self__)

    @classmethod
    async def delete_all(cls, filter: dict) -> None:
        await DataHub.delete_all(cls, filter)

    @property
    def data(self):
        obj_data = self.dict()

        return obj_data

    @classmethod
    async def get(cls, object_id: str) -> 'DataModel':
        return await DataHub.get(cls, object_id)

    @classmethod
    async def find(cls, filter: Union[DataFilter,dict]) -> List['DataModel']:
        if isinstance(filter, dict):
            filter = DataFilter(fields=filter)

        return await DataHub.find(cls, filter)



class TargetGetterProxy:
    def __init__(self, ctx, target):
        self.ctx = ctx
        self.target = target

    def get(self):
        return getattr(self.ctx, self.target)

    def __str__(self):
        return str(self.get())



from .target_getter import TargetGetterProxy


class ProxyMixin:
    @classmethod
    def class_proxy_target(cls, target):
        return TargetGetterProxy(cls, target)

    def self_proxy_target(self, target):
        return TargetGetterProxy(self, target)
 


from datetime import datetime

from pydantic import BaseModel

from .base_internal import BaseInternal


class BaseDataModel(BaseModel):
    class Internal(BaseInternal):
        pass

    object_id: str = None
    created_at: datetime
    updated_at: datetime




from typing import List

from pydantic import BaseModel

from eyja.db.rethinkdb import RethinkDBConfig
from eyja.db.redis import RedisConfig
from eyja.errors import LoadConfigError


class Config(BaseModel):
    rethinkdb: List[RethinkDBConfig] = []
    redis: List[RedisConfig] = []

    def __init__(self, **data: dict) -> None:
        super().__init__(**data)

        if len(list(filter(lambda item: item.name == 'default', self.rethinkdb))) > 1:
            raise LoadConfigError('If there is more than one RethinkDB configuration, then each must have a unique name.')

        if len(list(filter(lambda item: item.name == 'default', self.redis))) > 1:
            raise LoadConfigError('If there is more than one Redis configuration, then each must have a unique name.')


from slugify import slugify

from .data_model import DataModel
from .base_redis_internal import BaseRedisInternal


class RedisDataModel(DataModel):
    class Internal(BaseRedisInternal):
        pass

    @property
    def key(self):
        result = []

        key_template_nodes = self.Internal.key_template.split('.')
        for node in key_template_nodes:
            if hasattr(self, node):
                result.append(slugify(getattr(self, node)))
            else:
                result.append(node)

        return '.'.join(result)

    @classmethod
    def filtered_key(cls, filter = {}):
        result = []

        key_template_nodes = cls.Internal.key_template.split('.')
        for node in key_template_nodes:
            if node in filter:
                result.append(slugify(filter[node]))
            else:
                result.append('*')

        return '.'.join(result)



from eyja.db.rethinkdb import RethinkDBClient
from eyja.db.redis import RedisClient
from eyja.errors import ParseModelNamespaceError

from .config_hub import ConfigHub


class DataHub:
    _connections = {
        'rethinkdb': {},
        'redis': {},
    }

    @classmethod
    async def init(cls, config = None):
        if not config:
            await ConfigHub.load_from_file()
        else:
            ConfigHub.load(config)

        await cls.init_rethinkdb()
        await cls.init_redis()

    @classmethod
    async def init_rethinkdb(cls):
        for _config in ConfigHub.rethinkdb():
            cls._connections['rethinkdb'][_config.name] = RethinkDBClient(_config)
            await cls._connections['rethinkdb'][_config.name].init()

    @classmethod
    async def init_redis(cls):
        for _config in ConfigHub.redis():
            cls._connections['redis'][_config.name] = RedisClient(_config)
            await cls._connections['redis'][_config.name].init()

    @classmethod
    def get_connection(cls, obj):
        namespaces = str(obj.Internal.namespace).split(':')
        if len(namespaces) != 4:
            raise ParseModelNamespaceError(
                'A namespace must have four parts - storage type, storage connection, objectspace, and object type - separated by ":"'
            )

        storage_type = namespaces[0]
        storage_connection = namespaces[1]
        object_space = namespaces[2]
        object_type = namespaces[3]

        if len(storage_connection) < 1:
            storage_connection = 'default'

        if storage_type not in cls._connections:
            raise ParseModelNamespaceError(
                'Unknown storage type'
            )

        if storage_connection not in cls._connections[storage_type]:
            raise ParseModelNamespaceError(
                f'Unknown storage connection - [{storage_connection}]'
            )

        return (
            cls._connections[storage_type][storage_connection],
            object_space, object_type,
        )

    @classmethod
    async def save(cls, obj):
        connection, object_space, object_type = cls.get_connection(obj)
        await connection.save(obj, object_space, object_type)

    @classmethod
    async def delete(cls, obj):
        connection, object_space, object_type = cls.get_connection(obj)
        await connection.delete(obj, object_space, object_type)

    @classmethod
    async def delete_all(cls, obj, filter):
        connection, object_space, object_type = cls.get_connection(obj)
        await connection.delete_all(obj, object_space, object_type, filter)

    @classmethod
    async def get(cls, obj_cls, object_id):
        connection, object_space, object_type = cls.get_connection(obj_cls)
        return await connection.get(obj_cls, object_space, object_type, object_id)

    @classmethod
    async def find(cls, obj_cls, filter):
        connection, object_space, object_type = cls.get_connection(obj_cls)
        return await connection.find(obj_cls, object_space, object_type, filter)



import aioredis
import uuid
import json

from eyja.db.redis import RedisConfig
from eyja.errors import MissedRepresentationError
from eyja.utils import EyjaJSONEncoder


class RedisClient:
    _connection = {}

    def __init__(self, config: RedisConfig):
        self._config = config

    async def _connect(self):
        for db_name, db_id in self._config.db.items():
            self._connection[db_name] = aioredis.from_url(
                f'redis://{self._config.host}:{self._config.port}/{db_id}'
            )

    def get_redis_url(self, db_name):
        return f'redis://{self._config.host}:{self._config.port}/{self._config.db[db_name]}'

    def get_redis_connection(self, db_name):
        return aioredis.from_url(self.get_redis_url(db_name))

    async def init(self):
        pass

    async def save(self, obj, object_space, object_type):
        if 'key' not in dir(obj):
            raise MissedRepresentationError('Missed "key" representation')

        if not obj.object_id:
            obj.object_id = str(uuid.uuid4())

        hierarchical_key = f'{object_type}.{obj.key}'

        async with self.get_redis_connection(object_space).client() as conn:
            await conn.delete(hierarchical_key)

            set_args = [
                hierarchical_key,
                json.dumps(obj.data, cls=EyjaJSONEncoder),
            ]

            if obj.Internal.expiration > 0:
                set_args.append(obj.Internal.expiration)

            await conn.set(*set_args)

    async def delete(self, obj, object_space, object_type):
        if 'key' not in dir(obj):
            raise MissedRepresentationError('Missed "key" representation')

        hierarchical_key = f'{object_type}.{obj.key}'

        async with self.get_redis_connection(object_space).client() as conn:
            await conn.delete(hierarchical_key)

    async def get(self, obj_cls, object_space, object_type, object_id):
        from eyja.db.models import DataFilter

        filter = DataFilter(fields={'object_id': object_id})
        items = await self.find(obj_cls, object_space, object_type, filter)
        
        if len(items) > 0:
            return items[0]
        
        return None

    async def find(self, obj_cls, object_space, object_type, filter):
        result = []

        if 'filtered_key' not in dir(obj_cls):
            raise MissedRepresentationError('Missed "filtered_key" representation')

        search_key = obj_cls.filtered_key(filter.fields)
        hierarchical_key = f'{object_type}.{search_key}'
        async with self.get_redis_connection(object_space).client() as conn:
            keys = await conn.keys(hierarchical_key)
            for key in keys:
                data = await conn.get(key)
                obj_data = json.loads(data)
                result.append(obj_cls(**obj_data))
        
        return result